==> daemonshogi-0.1.3.tar.gz: daemonshogi-0.1.3/src/si/hash.c <==
/* SI SHOGI Library
   Copyright (C) 2002 Masahiko Tokita
   This file is part of the SI SHOGI Library.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.w

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with The Daemonshogi; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307 USA.  */
/*
 * hash.c
 * $Source: /cvs/si/hash.c,v $
 * $Id: hash.c,v 1.50 2002/09/18 10:33:56 mik Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "si.h"
#include "ui.h"

static const unsigned long MEM_SIZE= 64 * 1024 * 1024;
unsigned long HASH_MAX = 0x1FFFF;
si_int64 hashval[32][16 * 11] = {
#include "hashval.h"
};
static void *hash;

/**
 * ϥåơ֥Ѥ˥ݤ롣
 */
void newHASH(void) {
  hash = (void *)malloc( sizeof(HASH) * (HASH_MAX + 1));
  if (hash == NULL) {
    si_abort("No enough memory. In hash.c#newHASH()\n");
  }

  initHASH();
}

/**
 * ϥåơ֥롣
 */
void initHASH(void) {
  memset( (void *)hash, 0, sizeof(HASH) * HASH_MAX );
}

/** 
 * ϥåơ֥롣
 */
void freeHASH(void) {
  free(hash);
}

/**
 * BOARD ϥåͤ롣
 * @param bo оݤBOARD
 * @return int ϥå(24bit)
 */
si_int64 hsGetHashKey(BOARD *bo) {
  si_int64 key;
  extern const int arr_xy[];
  int i;

#ifdef DEBUG
  assert( bo != NULL );
#endif /* DEBUG */

  key = 0;

  if ( bo->next == SENTE ) {
    for ( i=1; i<=81; i++ ) {
      key ^= hashval[bo->board[ arr_xy[ i ] ]][ arr_xy[ i ] ];
    }
  } else {
    /* ֤ξΥϥå奭ϥӥåȿž */
    for ( i=1; i<=81; i++ ) {
      key ^= ( ~ hashval[bo->board[ arr_xy[ i ] ]][ arr_xy[ i ] ] );
    }
  }

  return key;
}

/**
 * BOARDϥåơ֥Ͽ롣
 * @param bo оݤBOARD
 * @param mate_stat ̤ɾ
 * @param count 
 * @param r_count м
 */
void hsPutBOARD(BOARD *bo, int mate_stat, int count, int r_count) {
  HASH *hs;
  si_int64 key;
  si_int64 offset;
  extern long count_entry_hash;
  extern long count_crash2;

#ifdef DEBUG
  assert( bo != NULL );
#endif /* DEBUG */

  key = bo->key;

  offset = key & HASH_MAX;

  hs = (HASH *)((char *)hash + (sizeof(HASH) * offset));

  if ((count + 1) < hs->count && hs->key == key) {
    /* Ʊ key ϿƤ */
    count_crash2++;

    if ( hsSuperiorBOARD(hs, bo) == 0 ) {
      /* 꾯ʤǵͤǤ */
      /* hs λ */
      hsSetPiece(hs, bo);
    }
    
    return;
  }

  count_entry_hash++;

  hs->key = key;
  hs->mate_stat = mate_stat;
  hsSetPiece( hs, bo );
  hs->count = count;
}

/**
 * ̤ point 
 * @param bo оݤBOARD
 * @param p pointǼѿΥݥ
 * @param count 
 * @param r_count м
 * @return ̤Ĥä 1 Ĥʤä 0 ֤
 */
int hsGetBOARD(BOARD *bo, int *mate_stat, int count, int r_count) {
  HASH *hs;
  si_int64 key;
  si_int64 offset;
  extern long G_COUNT_HASH_CHECK_MATE;
  /* extern long G_COUNT_HASH_NOT_CHECK_MATE; */

#ifdef DEBUG
  assert( bo != NULL );
#endif /* DEBUG */

  key = bo->key;
  offset = key & HASH_MAX;
  hs = (HASH *)((char *)hash + (sizeof(HASH) * offset));

  if (key != hs->key) {
    /* 㤦̤Ǥä */
    return 0;
  }
  
  if ( hs->mate_stat == CHECK_MATE && count <= hs->count ) {
    /* ̤ͥ۴طĴ٤ */
    if ( hsSuperiorBOARD(hs, bo) ) {
      *mate_stat = CHECK_MATE;
      G_COUNT_HASH_CHECK_MATE++;
      return 1;
    }
  }
  /*
    else if ( hs->point == NOT_CHECK_MATE && count <= hs->count ) {
    *mate_stat = NOT_CHECK_MATE;
    G_COUNT_HASH_NOT_CHECK_MATE++;
    return 1;
  }
  */
  else if ( hs->point == UNKNOWN_CHECK_MATE ) {
    /*  */
    *mate_stat = UNKNOWN_CHECK_MATE;
    return 1;
  }

  /* դʤä */
  return 0;
}

/**
 * min/maxѤBOARDϥåơ֥Ͽ롣
 * @param bo оݤBOARD
 * @param mate_stat ̤ɾ
 * @param count 
 * @param r_count м
 */
void hsPutBOARDMinMax(BOARD *bo, int point, int count, int r_count) {
  HASH *hs;
  si_int64 key;
  si_int64 offset;
  extern long count_entry_hash;

#ifdef DEBUG
  assert( bo != NULL );
#endif /* DEBUG */

  key = bo->key;
  offset = key & HASH_MAX;

  hs = (HASH *)((char *)hash + (sizeof(HASH) * offset));

  if ( ! ( hs->count <= count + 1 ) ) {
    /* Ʊ key ϿƤ */
    
    /* Ʊ̤ */
    if (key == hs->key) {
      /* 㤦̤äΤ롣ϤǺƥϥå夹롣 */
      return;
    }
    
    return;
  }

  count_entry_hash++;

  /*
  if ( point == 778 ) {
    putBOARD(bo);
  }
  */

  hs->key = key;
  hs->point = point;
  hsSetPiece( hs, bo );
  hs->count = count;
  /* hs->flguse = 1; */
}

/**
 * ̤ point min/maxѡ
 * @param bo оݤBOARD
 * @param p pointǼѿΥݥ
 * @param count 
 * @param r_count м
 * @return ̤Ĥä 1 Ĥʤä 0 ֤
 */
int hsGetBOARDMinMax(BOARD *bo, int *point, int count, int r_count) {
  HASH *hs;
  si_int64 key;
  si_int64 offset;

#ifdef DEBUG
  assert( bo != NULL );
#endif /* DEBUG */

  key = bo->key;
  offset = key & HASH_MAX;
  hs = (HASH *)((char *)hash + (sizeof(HASH) * offset));

  /*
  if ( hs->flguse == 0 )
    return 0;
  */
  
  if ( count <= hs->count ) {
    if ( bo->key != key ) {
      /* ϿƤ̤㤬ä */
      return 0;
    }

    if ( hsCmpPiece(hs, bo, bo->next) ) {
      /* ϿƤ̤㤬ä */
      /* printf("not same board2.\n"); */
      return 0;
    }

    /*
    if ( hs->point == 778 ) {
      printf("hsGetBOARDMinMax()\n");
      putBOARD(bo);
    }
     */ 
    
    *point = hs->point;

    return 1;
  }

  /* դʤä */
  return 0;
}

/**
 * HASH˻򥻥åȤ롣
 * @param hs оݤHASH
 * @param bo оݤBOARD
 */
void hsSetPiece(HASH *hs, BOARD *bo) {
  hs->piece[ SENTE ] =
    bo->piece[ SENTE ][ 1 ] +
    (bo->piece[ SENTE ][ 2 ] << 5) +
    (bo->piece[ SENTE ][ 3 ] << 8) +
    (bo->piece[ SENTE ][ 4 ] << 11) +
    (bo->piece[ SENTE ][ 5 ] << 14) +
    (bo->piece[ SENTE ][ 6 ] << 17) +
    (bo->piece[ SENTE ][ 7 ] << 19);
  hs->piece[ GOTE ] =
    bo->piece[ GOTE ][ 1 ] +
    (bo->piece[ GOTE ][ 2 ] << 5) +
    (bo->piece[ GOTE ][ 3 ] << 8) +
    (bo->piece[ GOTE ][ 4 ] << 11) +
    (bo->piece[ GOTE ][ 5 ] << 14) +
    (bo->piece[ GOTE ][ 6 ] << 17) +
    (bo->piece[ GOTE ][ 7 ] << 19);
}

/**
 * ̤ͥ۴طĴ٤롣
 * hs λ bo λʬä 1 ֤ʳ 0 ֤
 * hs ̤ǵͤʬäƤƳ bo Ʊ̳ hs λ
 * bo λʬä bo ϵͤࡣ
 * @param hs оݤHASH
 * @param bo оݤBOARD
 * @return bo λ hs λʬä 1 ֤
 *         ʳ 0 ֤
 */
int hsSuperiorBOARD(HASH *hs, BOARD *bo) {
  int next = bo->next;
  int p = hs->piece[ next ];

  if ( bo->piece[ next ][ FU ] < (p & 0x1f)) {
    /* ξ */
    return 0;
  }
  if ( bo->piece[ next ][ KYO ] < ((p >> 5) & 0x7)) {
    /* ξ */
    return 0;
  }
  if ( bo->piece[ next ][ KEI ] < ((p >> 8) & 0x7)) {
    /* ˤξ */
    return 0;
  }
  if ( bo->piece[ next ][ GIN ] < ((p >> 11) & 0x7)) {
    /* ξ */
    return 0;
  }
  if ( bo->piece[ next ][ GIN ] < ((p >> 14) & 0x7)) {
    /* ξ */
    return 0;
  }
  if ( bo->piece[ next ][ KAKU ] < ((p >> 17) & 0x3)) {
    /* Ѥξ */
    return 0;
  }
  if ( bo->piece[ next ][ HISYA ] < ((p >> 19) & 0x3)) {
    /* ξ */
    return 0;
  }
  
  return 1;
}

/**
 * ꤷ֤λӤ롣
 * Ʊʤ 0 ֤㤬 1 ֤
 * @param hs оݤHASH
 * @param bo оݤBOARD
 * @param next 
 * @return Ʊʤ 0 ֤㤬 1 ֤
 */
int hsCmpPiece(HASH *hs, BOARD *bo, int next) {
  int p = hs->piece[ next ];

  if ( bo->piece[ next ][ FU ] < (p & 0x1f)) {
    /* ξ */
    return 0;
  }
  if ( bo->piece[ next ][ KYO ] < ((p >> 5) & 0x7)) {
    /* ξ */
    return 0;
  }
  if ( bo->piece[ next ][ KEI ] < ((p >> 8) & 0x7)) {
    /* ˤξ */
    return 0;
  }
  if ( bo->piece[ next ][ GIN ] < ((p >> 11) & 0x7)) {
    /* ξ */
    return 0;
  }
  if ( bo->piece[ next ][ GIN ] < ((p >> 14) & 0x7)) {
    /* ξ */
    return 0;
  }
  if ( bo->piece[ next ][ KAKU ] < ((p >> 17) & 0x3)) {
    /* Ѥξ */
    return 0;
  }
  if ( bo->piece[ next ][ HISYA ] < ((p >> 19) & 0x3)) {
    /* ξ */
    return 0;
  }
  
  return 1;
}

/**
 * XYΰ֤Υϥåͤ­
 * @param bo оݤBOARD
 * @param xy ΰ
 */
void add_hash_value(BOARD *bo, int xy) {
  int p;
  
#ifdef DEBUG
  assert(bo != NULL);
  assert( 1 <= (xy >> 4) && (xy >> 4) <= 9 );
  assert( 1 <= (xy & 0xF) && (xy & 0xF) <= 9 );
#endif /* DEBUG */
  
  p = boGetPiece(bo, xy);

  if ( bo->next == SENTE ) {
    bo->key ^= hashval[ p ][ xy ];
  } else {
    bo->key ^= ( ~ hashval[ p ][ xy ] );
  }
}

/**
 * XYΰ֤Υϥåͤ򳰤
 * @param bo оݤBOARD
 * @param xy ΰ
 */
void sub_hash_value(BOARD *bo, int xy) {
  int p;
  
#ifdef DEBUG
  assert(bo != NULL);
  assert( 1 <= (xy >> 4) && (xy >> 4) <= 9 );
  assert( 1 <= (xy & 0xF) && (xy & 0xF) <= 9 );
#endif /* DEBUG */
  
  p = boGetPiece(bo, xy);
  
  if ( bo->next == SENTE ) {
    bo->key ^= hashval[ p ][ xy ];
  } else {
    bo->key ^= ( ~ hashval[ p ][ xy ] );
  }
}

==> daemonshogi-0.1.4.tar.gz: daemonshogi-0.1.4/src/si/hash.c <==
/* SI SHOGI Library
   Copyright (C) 2002 Masahiko Tokita
   This file is part of the SI SHOGI Library.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.w

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with The Daemonshogi; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307 USA.  */
/*
 * hash.c
 * $Source: /cvsroot/daemonshogi/daemonshogi/src/si/hash.c,v $
 * $Id: hash.c,v 1.3 2005/12/09 11:51:03 tokita Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "si.h"
#include "ui.h"

static const unsigned long MEM_SIZE= (2 * 1024 * 1024);
unsigned long HASH_MAX = 0x1FFFF;
#include "hashval.h"
static void *hash;

/**
 * ϥåơ֥Ѥ˥ݤ롣
 */
void newHASH(void) {
  hash = (void *)malloc( sizeof(HASH) * (HASH_MAX + 1));
  if (hash == NULL) {
    si_abort("No enough memory. In hash.c#newHASH()\n");
  }

  initHASH();
}

/**
 * ϥåơ֥롣
 */
void initHASH(void) {
  memset( (void *)hash, 0, sizeof(HASH) * HASH_MAX );
}

/** 
 * ϥåơ֥롣
 */
void freeHASH(void) {
  free(hash);
}

/**
 * BOARD ϥåͤ롣
 * @param bo оݤBOARD
 * @return int ϥå(24bit)
 */
si_int64 hsGetHashKey(BOARD *bo) {
  si_int64 key;
  extern const int arr_xy[];
  int i;

#ifdef DEBUG
  assert( bo != NULL );
#endif /* DEBUG */

  key = 0;

  if ( bo->next == SENTE ) {
    for ( i=1; i<=81; i++ ) {
      key ^= hashval[bo->board[ arr_xy[ i ] ]][ arr_xy[ i ] ];
    }
  } else {
    /* ֤ξΥϥå奭ϥӥåȿž */
    for ( i=1; i<=81; i++ ) {
      key ^= ( ~ hashval[bo->board[ arr_xy[ i ] ]][ arr_xy[ i ] ] );
    }
  }

  return key;
}

/**
 * BOARDϥåơ֥Ͽ롣
 * @param bo оݤBOARD
 * @param mate_stat ̤ɾ
 * @param count 
 * @param r_count м
 */
void hsPutBOARD(BOARD *bo, int mate_stat, int count, int r_count) {
  HASH *hs;
  si_int64 key;
  si_int64 offset;
  extern long count_entry_hash;
  extern long count_crash2;

#ifdef DEBUG
  assert( bo != NULL );
#endif /* DEBUG */

  key = bo->key;

  offset = key & HASH_MAX;

  hs = (HASH *)((char *)hash + (sizeof(HASH) * offset));

  if ((count + 1) < hs->count && hs->key == key) {
    /* Ʊ key ϿƤ */
    count_crash2++;

    if ( hsSuperiorBOARD(hs, bo) == 0 ) {
      /* 꾯ʤǵͤǤ */
      /* hs λ */
      hsSetPiece(hs, bo);
    }
    
    return;
  }

  count_entry_hash++;

  hs->key = key;
  hs->mate_stat = mate_stat;
  hsSetPiece( hs, bo );
  hs->count = count;
}

/**
 * ̤ point 
 * @param bo оݤBOARD
 * @param p pointǼѿΥݥ
 * @param count 
 * @param r_count м
 * @return ̤Ĥä 1 Ĥʤä 0 ֤
 */
int hsGetBOARD(BOARD *bo, int *mate_stat, int count, int r_count) {
  HASH *hs;
  si_int64 key;
  si_int64 offset;
  extern long G_COUNT_HASH_CHECK_MATE;
  /* extern long G_COUNT_HASH_NOT_CHECK_MATE; */

#ifdef DEBUG
  assert( bo != NULL );
#endif /* DEBUG */

  key = bo->key;
  offset = key & HASH_MAX;
  hs = (HASH *)((char *)hash + (sizeof(HASH) * offset));

  if (key != hs->key) {
    /* 㤦̤Ǥä */
    return 0;
  }
  
  if ( hs->mate_stat == CHECK_MATE && count <= hs->count ) {
    /* ̤ͥ۴طĴ٤ */
    if ( hsSuperiorBOARD(hs, bo) ) {
      *mate_stat = CHECK_MATE;
      G_COUNT_HASH_CHECK_MATE++;
      return 1;
    }
  }
  /*
    else if ( hs->point == NOT_CHECK_MATE && count <= hs->count ) {
    *mate_stat = NOT_CHECK_MATE;
    G_COUNT_HASH_NOT_CHECK_MATE++;
    return 1;
  }
  */
  else if ( hs->point == UNKNOWN_CHECK_MATE ) {
    /*  */
    *mate_stat = UNKNOWN_CHECK_MATE;
    return 1;
  }

  /* դʤä */
  return 0;
}

/**
 * min/maxѤBOARDϥåơ֥Ͽ롣
 * @param bo оݤBOARD
 * @param mate_stat ̤ɾ
 * @param count 
 * @param r_count м
 */
void hsPutBOARDMinMax(BOARD *bo, int point, int count, int r_count) {
  HASH *hs;
  si_int64 key;
  si_int64 offset;
  extern long count_entry_hash;

#ifdef DEBUG
  assert( bo != NULL );
#endif /* DEBUG */

  key = bo->key;
  offset = key & HASH_MAX;

  hs = (HASH *)((char *)hash + (sizeof(HASH) * offset));

  if ( ! ( hs->count <= count + 1 ) ) {
    /* Ʊ key ϿƤ */
    
    /* Ʊ̤ */
    if (key == hs->key) {
      /* 㤦̤äΤ롣ϤǺƥϥå夹롣 */
      return;
    }
    
    return;
  }

  count_entry_hash++;

  /*
  if ( point == 778 ) {
    putBOARD(bo);
  }
  */

  hs->key = key;
  hs->point = point;
  hsSetPiece( hs, bo );
  hs->count = count;
  /* hs->flguse = 1; */
}

/**
 * ̤ point min/maxѡ
 * @param bo оݤBOARD
 * @param p pointǼѿΥݥ
 * @param count 
 * @param r_count м
 * @return ̤Ĥä 1 Ĥʤä 0 ֤
 */
int hsGetBOARDMinMax(BOARD *bo, int *point, int count, int r_count) {
  HASH *hs;
  si_int64 key;
  si_int64 offset;

#ifdef DEBUG
  assert( bo != NULL );
#endif /* DEBUG */

  key = bo->key;
  offset = key & HASH_MAX;
  hs = (HASH *)((char *)hash + (sizeof(HASH) * offset));

  /*
  if ( hs->flguse == 0 )
    return 0;
  */
  
  if ( count <= hs->count ) {
    if ( bo->key != key ) {
      /* ϿƤ̤㤬ä */
      return 0;
    }

    if ( hsCmpPiece(hs, bo, bo->next) ) {
      /* ϿƤ̤㤬ä */
      /* printf("not same board2.\n"); */
      return 0;
    }

    /*
    if ( hs->point == 778 ) {
      printf("hsGetBOARDMinMax()\n");
      putBOARD(bo);
    }
     */ 
    
    *point = hs->point;

    return 1;
  }

  /* դʤä */
  return 0;
}

/**
 * HASH˻򥻥åȤ롣
 * @param hs оݤHASH
 * @param bo оݤBOARD
 */
void hsSetPiece(HASH *hs, BOARD *bo) {
  hs->piece[ SENTE ] =
    bo->piece[ SENTE ][ 1 ] +
    (bo->piece[ SENTE ][ 2 ] << 5) +
    (bo->piece[ SENTE ][ 3 ] << 8) +
    (bo->piece[ SENTE ][ 4 ] << 11) +
    (bo->piece[ SENTE ][ 5 ] << 14) +
    (bo->piece[ SENTE ][ 6 ] << 17) +
    (bo->piece[ SENTE ][ 7 ] << 19);
  hs->piece[ GOTE ] =
    bo->piece[ GOTE ][ 1 ] +
    (bo->piece[ GOTE ][ 2 ] << 5) +
    (bo->piece[ GOTE ][ 3 ] << 8) +
    (bo->piece[ GOTE ][ 4 ] << 11) +
    (bo->piece[ GOTE ][ 5 ] << 14) +
    (bo->piece[ GOTE ][ 6 ] << 17) +
    (bo->piece[ GOTE ][ 7 ] << 19);
}

/**
 * ̤ͥ۴طĴ٤롣
 * hs λ bo λʬä 1 ֤ʳ 0 ֤
 * hs ̤ǵͤʬäƤƳ bo Ʊ̳ hs λ
 * bo λʬä bo ϵͤࡣ
 * @param hs оݤHASH
 * @param bo оݤBOARD
 * @return bo λ hs λʬä 1 ֤
 *         ʳ 0 ֤
 */
int hsSuperiorBOARD(HASH *hs, BOARD *bo) {
  int next = bo->next;
  int p = hs->piece[ next ];

  if ( bo->piece[ next ][ FU ] < (p & 0x1f)) {
    /* ξ */
    return 0;
  }
  if ( bo->piece[ next ][ KYO ] < ((p >> 5) & 0x7)) {
    /* ξ */
    return 0;
  }
  if ( bo->piece[ next ][ KEI ] < ((p >> 8) & 0x7)) {
    /* ˤξ */
    return 0;
  }
  if ( bo->piece[ next ][ GIN ] < ((p >> 11) & 0x7)) {
    /* ξ */
    return 0;
  }
  if ( bo->piece[ next ][ GIN ] < ((p >> 14) & 0x7)) {
    /* ξ */
    return 0;
  }
  if ( bo->piece[ next ][ KAKU ] < ((p >> 17) & 0x3)) {
    /* Ѥξ */
    return 0;
  }
  if ( bo->piece[ next ][ HISYA ] < ((p >> 19) & 0x3)) {
    /* ξ */
    return 0;
  }
  
  return 1;
}

/**
 * ꤷ֤λӤ롣
 * Ʊʤ 0 ֤㤬 1 ֤
 * @param hs оݤHASH
 * @param bo оݤBOARD
 * @param next 
 * @return Ʊʤ 0 ֤㤬 1 ֤
 */
int hsCmpPiece(HASH *hs, BOARD *bo, int next) {
  int p = hs->piece[ next ];

  if ( bo->piece[ next ][ FU ] < (p & 0x1f)) {
    /* ξ */
    return 0;
  }
  if ( bo->piece[ next ][ KYO ] < ((p >> 5) & 0x7)) {
    /* ξ */
    return 0;
  }
  if ( bo->piece[ next ][ KEI ] < ((p >> 8) & 0x7)) {
    /* ˤξ */
    return 0;
  }
  if ( bo->piece[ next ][ GIN ] < ((p >> 11) & 0x7)) {
    /* ξ */
    return 0;
  }
  if ( bo->piece[ next ][ GIN ] < ((p >> 14) & 0x7)) {
    /* ξ */
    return 0;
  }
  if ( bo->piece[ next ][ KAKU ] < ((p >> 17) & 0x3)) {
    /* Ѥξ */
    return 0;
  }
  if ( bo->piece[ next ][ HISYA ] < ((p >> 19) & 0x3)) {
    /* ξ */
    return 0;
  }
  
  return 1;
}

/**
 * XYΰ֤Υϥåͤ­
 * @param bo оݤBOARD
 * @param xy ΰ
 */
void add_hash_value(BOARD *bo, int xy) {
  int p;
  
#ifdef DEBUG
  assert(bo != NULL);
  assert( 1 <= (xy >> 4) && (xy >> 4) <= 9 );
  assert( 1 <= (xy & 0xF) && (xy & 0xF) <= 9 );
#endif /* DEBUG */
  
  p = boGetPiece(bo, xy);

  if ( bo->next == SENTE ) {
    bo->key ^= hashval[ p ][ xy ];
  } else {
    bo->key ^= ( ~ hashval[ p ][ xy ] );
  }
}

/**
 * XYΰ֤Υϥåͤ򳰤
 * @param bo оݤBOARD
 * @param xy ΰ
 */
void sub_hash_value(BOARD *bo, int xy) {
  int p;
  
#ifdef DEBUG
  assert(bo != NULL);
  assert( 1 <= (xy >> 4) && (xy >> 4) <= 9 );
  assert( 1 <= (xy & 0xF) && (xy & 0xF) <= 9 );
#endif /* DEBUG */
  
  p = boGetPiece(bo, xy);
  
  if ( bo->next == SENTE ) {
    bo->key ^= hashval[ p ][ xy ];
  } else {
    bo->key ^= ( ~ hashval[ p ][ xy ] );
  }
}
